@@ -37,12 +37,14 @@
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
+#include <bluetooth/sdp.h>
+#include <bluetooth/sdp_lib.h>
 #include <bluetooth/mgmt.h>
 
 #include "plugin.h"
 #include "log.h"
-#include "manager.h"
 #include "adapter.h"
+#include "manager.h"
 #include "device.h"
 #include "event.h"
 
@@ -55,11 +57,15 @@ static struct controller_info {
 	uint8_t type;
 	bdaddr_t bdaddr;
 	uint8_t features[8];
+	uint8_t dev_class[3];
 	uint16_t manufacturer;
 	uint8_t hci_ver;
 	uint16_t hci_rev;
 	gboolean enabled;
-	uint8_t mode;
+	gboolean connectable;
+	gboolean discoverable;
+	gboolean pairable;
+	uint8_t sec_mode;
 } *controllers = NULL;
 
 static int mgmt_sock = -1;
@@ -84,7 +90,7 @@ static void read_version_complete(int sk, void *buf, size_t len)
 	DBG("version %u revision %u", mgmt_version, mgmt_revision);
 
 	memset(&hdr, 0, sizeof(hdr));
-	hdr.opcode = MGMT_OP_READ_INDEX_LIST;
+	hdr.opcode = htobs(MGMT_OP_READ_INDEX_LIST);
 	if (write(sk, &hdr, sizeof(hdr)) < 0)
 		error("Unable to read controller index list: %s (%d)",
 						strerror(errno), errno);
@@ -112,7 +118,7 @@ static void read_info(int sk, uint16_t index)
 	struct mgmt_cp_read_info *cp = (void *) &buf[sizeof(*hdr)];
 
 	memset(buf, 0, sizeof(buf));
-	hdr->opcode = MGMT_OP_READ_INFO;
+	hdr->opcode = htobs(MGMT_OP_READ_INFO);
 	hdr->len = htobs(sizeof(*cp));
 
 	cp->index = htobs(index);
@@ -146,7 +152,7 @@ static void remove_controller(uint16_t index)
 	if (!controllers[index].valid)
 		return;
 
-	manager_unregister_adapter(index);
+	btd_manager_unregister_adapter(index);
 
 	memset(&controllers[index], 0, sizeof(struct controller_info));
 
@@ -168,21 +174,299 @@ static void mgmt_index_removed(int sk, void *buf, size_t len)
 	remove_controller(index);
 }
 
-static void read_mode(int sk, uint16_t index)
+static int mgmt_set_mode(int index, uint16_t opcode, uint8_t val)
 {
-	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_read_mode)];
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_mode)];
 	struct mgmt_hdr *hdr = (void *) buf;
-	struct mgmt_cp_read_mode *cp = (void *) &buf[sizeof(*hdr)];
+	struct mgmt_mode *cp = (void *) &buf[sizeof(*hdr)];
 
 	memset(buf, 0, sizeof(buf));
-	hdr->opcode = MGMT_OP_READ_MODE;
+	hdr->opcode = htobs(opcode);
 	hdr->len = htobs(sizeof(*cp));
 
 	cp->index = htobs(index);
+	cp->val = val;
 
-	if (write(sk, buf, sizeof(buf)) < 0)
-		error("Unable to send read_mode command: %s (%d)",
-						strerror(errno), errno);
+	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
+		return -errno;
+
+	return 0;
+}
+
+static int mgmt_set_connectable(int index, gboolean connectable)
+{
+	DBG("index %d connectable %d", index, connectable);
+	return mgmt_set_mode(index, MGMT_OP_SET_CONNECTABLE, connectable);
+}
+
+static int mgmt_set_discoverable(int index, gboolean discoverable)
+{
+	DBG("index %d discoverable %d", index, discoverable);
+	return mgmt_set_mode(index, MGMT_OP_SET_DISCOVERABLE, discoverable);
+}
+
+static int mgmt_set_pairable(int index, gboolean pairable)
+{
+	DBG("index %d pairable %d", index, pairable);
+	return mgmt_set_mode(index, MGMT_OP_SET_PAIRABLE, pairable);
+}
+
+static int mgmt_update_powered(int index, uint8_t powered)
+{
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	gboolean pairable, discoverable;
+	uint8_t on_mode;
+
+	if (index > max_index) {
+		error("Unexpected index %u", index);
+		return -ENODEV;
+	}
+
+	info = &controllers[index];
+
+	info->enabled = powered;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (adapter == NULL) {
+		DBG("Adapter not found");
+		return -ENODEV;
+	}
+
+	if (!powered) {
+		info->connectable = FALSE;
+		info->pairable = FALSE;
+		info->discoverable = FALSE;
+
+		btd_adapter_stop(adapter);
+		return 0;
+	}
+
+	btd_adapter_start(adapter);
+
+	btd_adapter_get_mode(adapter, NULL, &on_mode, &pairable);
+
+	discoverable = (on_mode == MODE_DISCOVERABLE);
+
+	if (on_mode == MODE_DISCOVERABLE && !info->discoverable)
+		mgmt_set_discoverable(index, TRUE);
+	else if (on_mode == MODE_CONNECTABLE && !info->connectable)
+		mgmt_set_connectable(index, TRUE);
+	else {
+		uint8_t mode = 0;
+
+		if (info->connectable)
+			mode |= SCAN_PAGE;
+		if (info->discoverable)
+			mode |= SCAN_INQUIRY;
+
+		adapter_mode_changed(adapter, mode);
+	}
+
+	if (info->pairable != pairable)
+		mgmt_set_pairable(index, pairable);
+
+	return 0;
+}
+
+static void mgmt_powered(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	uint16_t index;
+
+	if (len < sizeof(*ev)) {
+		error("Too small powered event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u powered %u", index, ev->val);
+
+	mgmt_update_powered(index, ev->val);
+}
+
+static void mgmt_discoverable(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+	uint8_t mode;
+
+	if (len < sizeof(*ev)) {
+		error("Too small discoverable event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u discoverable %u", index, ev->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in discoverable event", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->discoverable = ev->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	if (info->connectable)
+		mode = SCAN_PAGE;
+	else
+		mode = 0;
+
+	if (info->discoverable)
+		mode |= SCAN_INQUIRY;
+
+	adapter_mode_changed(adapter, mode);
+}
+
+static void mgmt_connectable(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+	uint8_t mode;
+
+	if (len < sizeof(*ev)) {
+		error("Too small connectable event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u connectable %u", index, ev->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in connectable event", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->connectable = ev->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	if (info->discoverable)
+		mode = SCAN_INQUIRY;
+	else
+		mode = 0;
+
+	if (info->connectable)
+		mode |= SCAN_PAGE;
+
+	adapter_mode_changed(adapter, mode);
+}
+
+static void mgmt_pairable(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *ev = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+
+	if (len < sizeof(*ev)) {
+		error("Too small pairable event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&ev->index));
+
+	DBG("Controller %u pairable %u", index, ev->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in pairable event", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->pairable = ev->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	btd_adapter_pairable_changed(adapter, info->pairable);
+}
+
+static void uuid_to_uuid128(uuid_t *uuid128, const uuid_t *uuid)
+{
+	if (uuid->type == SDP_UUID16)
+		sdp_uuid16_to_uuid128(uuid128, uuid);
+	else if (uuid->type == SDP_UUID32)
+		sdp_uuid32_to_uuid128(uuid128, uuid);
+	else
+		memcpy(uuid128, uuid, sizeof(*uuid));
+}
+
+static int mgmt_add_uuid(int index, uuid_t *uuid, uint8_t svc_hint)
+{
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_add_uuid)];
+	struct mgmt_hdr *hdr = (void *) buf;
+	struct mgmt_cp_add_uuid *cp = (void *) &buf[sizeof(*hdr)];
+	uuid_t uuid128;
+
+	DBG("index %d", index);
+
+	uuid_to_uuid128(&uuid128, uuid);
+
+	memset(buf, 0, sizeof(buf));
+	hdr->opcode = htobs(MGMT_OP_ADD_UUID);
+	hdr->len = htobs(sizeof(*cp));
+
+	cp->index = htobs(index);
+	memcpy(cp->uuid, uuid128.value.uuid128.data, 16);
+	cp->svc_hint = svc_hint;
+
+	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
+		return -errno;
+
+	return 0;
+}
+
+static int mgmt_remove_uuid(int index, uuid_t *uuid)
+{
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_remove_uuid)];
+	struct mgmt_hdr *hdr = (void *) buf;
+	struct mgmt_cp_remove_uuid *cp = (void *) &buf[sizeof(*hdr)];
+	uuid_t uuid128;
+
+	DBG("index %d", index);
+
+	uuid_to_uuid128(&uuid128, uuid);
+
+	memset(buf, 0, sizeof(buf));
+	hdr->opcode = htobs(MGMT_OP_REMOVE_UUID);
+	hdr->len = htobs(sizeof(*cp));
+
+	cp->index = htobs(index);
+	memcpy(cp->uuid, uuid128.value.uuid128.data, 16);
+
+	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
+		return -errno;
+
+	return 0;
+}
+
+static int clear_uuids(int index)
+{
+	uuid_t uuid_any;
+
+	memset(&uuid_any, 0, sizeof(uuid_any));
+	uuid_any.type = SDP_UUID128;
+
+	return mgmt_remove_uuid(index, &uuid_any);
 }
 
 static void read_index_list_complete(int sk, void *buf, size_t len)
@@ -210,13 +494,22 @@ static void read_index_list_complete(int sk, void *buf, size_t len)
 
 		add_controller(index);
 		read_info(sk, index);
+		clear_uuids(index);
 	}
 }
 
+static int mgmt_set_powered(int index, gboolean powered)
+{
+	DBG("index %d powered %d", index, powered);
+	return mgmt_set_mode(index, MGMT_OP_SET_POWERED, powered);
+}
+
 static void read_info_complete(int sk, void *buf, size_t len)
 {
 	struct mgmt_rp_read_info *rp = buf;
 	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint8_t mode;
 	uint16_t index;
 	char addr[18];
 
@@ -225,68 +518,174 @@ static void read_info_complete(int sk, void *buf, size_t len)
 		return;
 	}
 
-	if (rp->status != 0) {
-		error("Reading controller info failed: %s (%u)",
-					strerror(rp->status), rp->status);
-		return;
-	}
-
 	index = btohs(bt_get_unaligned(&rp->index));
 	if (index > max_index) {
 		error("Unexpected index %u in read info complete", index);
 		return;
 	}
 
+	mgmt_set_mode(index, MGMT_OP_SET_SERVICE_CACHE, 1);
+
 	info = &controllers[index];
 	info->type = rp->type;
+	info->enabled = rp->powered;
+	info->connectable = rp->connectable;
+	info->discoverable = rp->discoverable;
+	info->pairable = rp->pairable;
+	info->sec_mode = rp->sec_mode;
 	bacpy(&info->bdaddr, &rp->bdaddr);
+	memcpy(info->dev_class, rp->dev_class, 3);
 	memcpy(info->features, rp->features, 8);
 	info->manufacturer = btohs(bt_get_unaligned(&rp->manufacturer));
 	info->hci_ver = rp->hci_ver;
 	info->hci_rev = btohs(bt_get_unaligned(&rp->hci_rev));
 
 	ba2str(&info->bdaddr, addr);
-	DBG("hci%u addr %s type %u manufacturer %d hci ver %d:%d",
-				index, addr, info->type, info->manufacturer,
-				info->hci_ver, info->hci_rev);
+	DBG("hci%u type %u addr %s", index, info->type, addr);
+	DBG("hci%u class 0x%02x%02x%02x", index,
+		info->dev_class[2], info->dev_class[1], info->dev_class[0]);
+	DBG("hci%u manufacturer %d HCI ver %d:%d", index, info->manufacturer,
+						info->hci_ver, info->hci_rev);
+	DBG("hci%u enabled %u discoverable %u pairable %u sec_mode %u", index,
+					info->enabled, info->discoverable,
+					info->pairable, info->sec_mode);
 
-	read_mode(sk, index);
-}
-
-static void read_mode_complete(int sk, void *buf, size_t len)
-{
-	struct mgmt_rp_read_mode *rp = buf;
-	struct controller_info *info;
-	uint16_t index;
-
-	if (len < sizeof(*rp)) {
-		error("Too small read mode complete event (%zu != %zu)",
-							len, sizeof(*rp));
+	adapter = btd_manager_register_adapter(index);
+	if (adapter == NULL) {
+		error("mgmtops: unable to register adapter");
 		return;
 	}
 
-	if (rp->status != 0) {
-		error("Reading controller mode failed: %s (%u)",
-					strerror(rp->status), rp->status);
+	btd_adapter_get_mode(adapter, &mode, NULL, NULL);
+	if (mode == MODE_OFF) {
+		mgmt_set_powered(index, FALSE);
+		return;
+	}
+
+	if (info->enabled)
+		mgmt_update_powered(index, TRUE);
+	else
+		mgmt_set_powered(index, TRUE);
+
+	btd_adapter_unref(adapter);
+}
+
+static void set_powered_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	uint16_t index;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set powered complete event");
 		return;
 	}
 
 	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d powered %u", index, rp->val);
+
+	mgmt_update_powered(index, rp->val);
+}
+
+static void set_discoverable_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+	uint8_t mode;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set discoverable complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d discoverable %u", index, rp->val);
+
 	if (index > max_index) {
-		error("Unexpected index %u in read mode complete", index);
+		error("Unexpected index %u in discoverable complete", index);
 		return;
 	}
 
 	info = &controllers[index];
-	info->enabled = rp->enabled ? TRUE : FALSE;
-	info->mode = rp->mode;
 
-	DBG("hci%u enabled %u mode %u", index, info->enabled, info->mode);
+	info->discoverable = rp->val ? TRUE : FALSE;
 
-	manager_register_adapter(index, info->enabled);
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
 
-	if (info->enabled)
-		manager_start_adapter(index);
+	/* set_discoverable will always also change page scanning */
+	mode = SCAN_PAGE;
+
+	if (info->discoverable)
+		mode |= SCAN_INQUIRY;
+
+	adapter_mode_changed(adapter, mode);
+}
+
+static void set_connectable_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set connectable complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d connectable %u", index, rp->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in connectable complete", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->connectable = rp->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (adapter)
+		adapter_mode_changed(adapter, rp->val ? SCAN_PAGE : 0);
+}
+
+static void set_pairable_complete(int sk, void *buf, size_t len)
+{
+	struct mgmt_mode *rp = buf;
+	struct controller_info *info;
+	struct btd_adapter *adapter;
+	uint16_t index;
+
+	if (len < sizeof(*rp)) {
+		error("Too small set pairable complete event");
+		return;
+	}
+
+	index = btohs(bt_get_unaligned(&rp->index));
+
+	DBG("hci%d pairable %u", index, rp->val);
+
+	if (index > max_index) {
+		error("Unexpected index %u in pairable complete", index);
+		return;
+	}
+
+	info = &controllers[index];
+
+	info->pairable = rp->val ? TRUE : FALSE;
+
+	adapter = manager_find_adapter(&info->bdaddr);
+	if (!adapter)
+		return;
+
+	btd_adapter_pairable_changed(adapter, info->pairable);
 }
 
 static void mgmt_cmd_complete(int sk, void *buf, size_t len)
@@ -313,8 +712,29 @@ static void mgmt_cmd_complete(int sk, void *buf, size_t len)
 	case MGMT_OP_READ_INFO:
 		read_info_complete(sk, ev->data, len - sizeof(*ev));
 		break;
-	case MGMT_OP_READ_MODE:
-		read_mode_complete(sk, ev->data, len - sizeof(*ev));
+	case MGMT_OP_SET_POWERED:
+		set_powered_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_SET_DISCOVERABLE:
+		set_discoverable_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_SET_CONNECTABLE:
+		set_connectable_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_SET_PAIRABLE:
+		set_pairable_complete(sk, ev->data, len - sizeof(*ev));
+		break;
+	case MGMT_OP_ADD_UUID:
+		DBG("add_uuid complete");
+		break;
+	case MGMT_OP_REMOVE_UUID:
+		DBG("remove_uuid complete");
+		break;
+	case MGMT_OP_SET_DEV_CLASS:
+		DBG("set_dev_class complete");
+		break;
+	case MGMT_OP_SET_SERVICE_CACHE:
+		DBG("set_service_cache complete");
 		break;
 	default:
 		error("Unknown command complete for opcode %u", opcode);
@@ -410,6 +830,18 @@ static gboolean mgmt_event(GIOChannel *io, GIOCondition cond, gpointer user_data
 	case MGMT_EV_INDEX_REMOVED:
 		mgmt_index_removed(sk, buf + MGMT_HDR_SIZE, len);
 		break;
+	case MGMT_EV_POWERED:
+		mgmt_powered(sk, buf + MGMT_HDR_SIZE, len);
+		break;
+	case MGMT_EV_DISCOVERABLE:
+		mgmt_discoverable(sk, buf + MGMT_HDR_SIZE, len);
+		break;
+	case MGMT_EV_CONNECTABLE:
+		mgmt_connectable(sk, buf + MGMT_HDR_SIZE, len);
+		break;
+	case MGMT_EV_PAIRABLE:
+		mgmt_pairable(sk, buf + MGMT_HDR_SIZE, len);
+		break;
 	default:
 		error("Unknown Management opcode %u", opcode);
 		break;
@@ -441,7 +873,7 @@ static int mgmt_setup(void)
 	}
 
 	memset(&hdr, 0, sizeof(hdr));
-	hdr.opcode = MGMT_OP_READ_VERSION;
+	hdr.opcode = htobs(MGMT_OP_READ_VERSION);
 	if (write(dd, &hdr, sizeof(hdr)) < 0) {
 		err = -errno;
 		goto fail;
@@ -480,46 +912,31 @@ static void mgmt_cleanup(void)
 	}
 }
 
-static int mgmt_start(int index)
+static int mgmt_set_dev_class(int index, uint8_t major, uint8_t minor)
 {
-	DBG("index %d", index);
-	return -ENOSYS;
+	char buf[MGMT_HDR_SIZE + sizeof(struct mgmt_cp_set_dev_class)];
+	struct mgmt_hdr *hdr = (void *) buf;
+	struct mgmt_cp_set_dev_class *cp = (void *) &buf[sizeof(*hdr)];
+
+	DBG("index %d major %u minor %u", index, major, minor);
+
+	memset(buf, 0, sizeof(buf));
+	hdr->opcode = htobs(MGMT_OP_SET_DEV_CLASS);
+	hdr->len = htobs(sizeof(*cp));
+
+	cp->index = htobs(index);
+	cp->major = major;
+	cp->minor = minor;
+
+	if (write(mgmt_sock, buf, sizeof(buf)) < 0)
+		return -errno;
+
+	return 0;
 }
 
-static int mgmt_stop(int index)
+static int mgmt_set_limited_discoverable(int index, gboolean limited)
 {
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
-static int mgmt_powered(int index, gboolean powered)
-{
-	DBG("index %d powered %d", index, powered);
-	return -ENOSYS;
-}
-
-static int mgmt_connectable(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
-static int mgmt_discoverable(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
-static int mgmt_set_class(int index, uint32_t class)
-{
-	DBG("index %d class %u", index, class);
-	return -ENOSYS;
-}
-
-static int mgmt_set_limited_discoverable(int index, uint32_t class,
-							gboolean limited)
-{
-	DBG("index %d class %u, limited %d", index, class, limited);
+	DBG("index %d limited %d", index, limited);
 	return -ENOSYS;
 }
 
@@ -579,30 +996,18 @@ static int mgmt_fast_connectable(int index, gboolean enable)
 	return -ENOSYS;
 }
 
-static int mgmt_read_clock(int index, int handle, int which, int timeout,
+static int mgmt_read_clock(int index, bdaddr_t *bdaddr, int which, int timeout,
 					uint32_t *clock, uint16_t *accuracy)
-{
-	DBG("index %d handle %d which %d timeout %d", index, handle,
-							which, timeout);
-	return -ENOSYS;
-}
-
-static int mgmt_conn_handle(int index, const bdaddr_t *bdaddr, int *handle)
 {
 	char addr[18];
 
 	ba2str(bdaddr, addr);
-	DBG("index %d addr %s", index, addr);
+	DBG("index %d addr %s which %d timeout %d", index, addr, which,
+								timeout);
 
 	return -ENOSYS;
 }
 
-static int mgmt_write_eir_data(int index, uint8_t *data)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
 static int mgmt_read_bdaddr(int index, bdaddr_t *bdaddr)
 {
 	char addr[18];
@@ -676,18 +1081,6 @@ static int mgmt_read_local_features(int index, uint8_t *features)
 	return 0;
 }
 
-static int mgmt_read_local_ext_features(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
-static int mgmt_read_link_policy(int index)
-{
-	DBG("index %d", index);
-	return -ENOSYS;
-}
-
 static int mgmt_disconnect(int index, uint16_t handle)
 {
 	DBG("index %d handle %u", index, handle);
@@ -756,16 +1149,9 @@ static int mgmt_read_scan_enable(int index)
 	return -ENOSYS;
 }
 
-static int mgmt_write_le_host(int index, uint8_t le, uint8_t simul)
+static int mgmt_enable_le(int index)
 {
-	DBG("index %d le %u simul %u", index, le, simul);
-	return -ENOSYS;
-}
-
-static int mgmt_get_remote_version(int index, uint16_t handle,
-							gboolean delayed)
-{
-	DBG("index %d handle %u delayed %d", index, handle, delayed);
+	DBG("index %d", index);
 	return -ENOSYS;
 }
 
@@ -780,14 +1166,39 @@ static int mgmt_encrypt_link(int index, bdaddr_t *dst, bt_hci_result_t cb,
 	return -ENOSYS;
 }
 
+static int mgmt_set_did(int index, uint16_t vendor, uint16_t product,
+							uint16_t version)
+{
+	DBG("index %d vendor %u product %u version %u",
+					index, vendor, product, version);
+	return -ENOSYS;
+}
+
+static int mgmt_disable_cod_cache(int index)
+{
+	DBG("index %d", index);
+	return mgmt_set_mode(index, MGMT_OP_SET_SERVICE_CACHE, 0);
+}
+
+static int mgmt_restore_powered(int index)
+{
+	DBG("index %d", index);
+	return -ENOSYS;
+}
+
+static int mgmt_load_keys(int index, GSList *keys, gboolean debug_keys)
+{
+	DBG("index %d keys %d debug_keys %d", index, g_slist_length(keys),
+								debug_keys);
+	return -ENOSYS;
+}
+
 static struct btd_adapter_ops mgmt_ops = {
 	.setup = mgmt_setup,
 	.cleanup = mgmt_cleanup,
-	.start = mgmt_start,
-	.stop = mgmt_stop,
-	.set_powered = mgmt_powered,
-	.set_connectable = mgmt_connectable,
-	.set_discoverable = mgmt_discoverable,
+	.set_powered = mgmt_set_powered,
+	.set_discoverable = mgmt_set_discoverable,
+	.set_pairable = mgmt_set_pairable,
 	.set_limited_discoverable = mgmt_set_limited_discoverable,
 	.start_inquiry = mgmt_start_inquiry,
 	.stop_inquiry = mgmt_stop_inquiry,
@@ -796,19 +1207,15 @@ static struct btd_adapter_ops mgmt_ops = {
 	.resolve_name = mgmt_resolve_name,
 	.cancel_resolve_name = mgmt_cancel_resolve_name,
 	.set_name = mgmt_set_name,
-	.set_class = mgmt_set_class,
+	.set_dev_class = mgmt_set_dev_class,
 	.set_fast_connectable = mgmt_fast_connectable,
 	.read_clock = mgmt_read_clock,
-	.get_conn_handle = mgmt_conn_handle,
-	.write_eir_data = mgmt_write_eir_data,
 	.read_bdaddr = mgmt_read_bdaddr,
 	.block_device = mgmt_block_device,
 	.unblock_device = mgmt_unblock_device,
 	.get_conn_list = mgmt_get_conn_list,
 	.read_local_version = mgmt_read_local_version,
 	.read_local_features = mgmt_read_local_features,
-	.read_local_ext_features = mgmt_read_local_ext_features,
-	.read_link_policy = mgmt_read_link_policy,
 	.disconnect = mgmt_disconnect,
 	.remove_bonding = mgmt_remove_bonding,
 	.request_authentication = mgmt_request_authentication,
@@ -817,9 +1224,14 @@ static struct btd_adapter_ops mgmt_ops = {
 	.passkey_reply = mgmt_passkey_reply,
 	.get_auth_info = mgmt_get_auth_info,
 	.read_scan_enable = mgmt_read_scan_enable,
-	.write_le_host = mgmt_write_le_host,
-	.get_remote_version = mgmt_get_remote_version,
+	.enable_le = mgmt_enable_le,
 	.encrypt_link = mgmt_encrypt_link,
+	.set_did = mgmt_set_did,
+	.add_uuid = mgmt_add_uuid,
+	.remove_uuid = mgmt_remove_uuid,
+	.disable_cod_cache = mgmt_disable_cod_cache,
+	.restore_powered = mgmt_restore_powered,
+	.load_keys = mgmt_load_keys,
 };
 
 static int mgmt_init(void)
